from contextlib import contextmanager
import time


@contextmanager
def time_measure(name: str, enable: bool = None):
    start = time.time()
    yield
    if enable:
        print(f"name: {name}, cost: {time.time() - start}")


def unused_dict_to_list(cb_dict: dict) -> list:
    if not cb_dict:
        return []
    # by default, xml parser will create a dict, which only contains one element
    assert len(cb_dict) == 1
    ret = list(cb_dict.values())[0]
    if isinstance(ret, list):
        return ret
    return [
        ret,
    ]



import gzip
import io
import typing
import pathlib
import xmltodict

from cobertura_parser.models.builtin import CoberturaStructure
from cobertura_parser.ext.jacoco import jacoco2cobertura


class CoberturaLoader(object):
    """
    load data from outside
    static class which will not save anything because data can be huge
    """

    @classmethod
    def from_file(
        cls, file_path: typing.Union[str, pathlib.Path], *args, **kwargs
    ) -> typing.Union[CoberturaStructure, dict]:
        with open(file_path, encoding="utf-8") as f:
            return cls.from_str(f.read(), *args, **kwargs)

    @classmethod
    def from_str(
        cls, xml_content: str, to_dict: bool = None
    ) -> typing.Union[CoberturaStructure, dict]:
        d = xmltodict.parse(
            gzip.GzipFile(fileobj=io.BytesIO(gzip.compress(xml_content.encode())))
        )
        if to_dict:
            return d
        return CoberturaStructure(**d)

    @classmethod
    def from_jacoco_file(
        cls, file_path: typing.Union[str, pathlib.Path], *args, **kwargs
    ):
        return cls.from_str(jacoco2cobertura(file_path), *args, **kwargs)



import typing
from pydantic import BaseModel, validator
from cobertura_parser.models.builtin import (
    CoberturaMethod,
    CoberturaKlass,
    CoberturaPackage,
    CoberturaLine,
    CoberturaStructureSlim,
    CoberturaStructure,
)


class CodeSnapshotMethod(CoberturaMethod):
    _TYPE_LINE_FINAL = typing.Union[typing.List[int], typing.List[CoberturaLine]]
    lines: _TYPE_LINE_FINAL = None

    @validator("lines")
    def line2int(cls, lines: typing.List[CoberturaLine]):
        return [each.number for each in lines]


class CodeSnapshotKlass(CoberturaKlass):
    _TYPE_LINE_FINAL = typing.Union[typing.List[int], typing.List[CoberturaLine]]
    lines: _TYPE_LINE_FINAL = None
    methods: typing.List[CodeSnapshotMethod]

    @validator("lines")
    def line2int(cls, lines: typing.List[CoberturaLine]):
        return [each.number for each in lines]


class CodeSnapshotPackage(CoberturaPackage):
    classes: typing.List[CodeSnapshotKlass]


class CodeSnapshot(BaseModel):
    """snapshot does not need other attrs, so inherits from BaseModel"""

    sources: typing.Dict[str, typing.Union[str, typing.List[str]]] = None
    packages: typing.List[CodeSnapshotPackage]

    @classmethod
    def from_slim(cls, slim_data: CoberturaStructureSlim) -> "CodeSnapshot":
        return cls(**slim_data.dict())

    @classmethod
    def from_normal(cls, data: CoberturaStructure) -> "CodeSnapshot":
        return cls.from_slim(data.slim())


class CodeSnapshotExt(BaseModel):
    pass


class CodeSnapshotFat(BaseModel):
    data: CodeSnapshot
    extras: CodeSnapshotExt = dict()

    @classmethod
    def from_slim(cls, slim_data: CoberturaStructureSlim) -> "CodeSnapshotFat":
        # todo: extras will be auto calculated inside
        return cls(data=slim_data.dict())

    @classmethod
    def from_normal(cls, data: CoberturaStructure) -> "CodeSnapshotFat":
        return cls.from_slim(data.slim())



# from: https://github.com/rix0rrr/cover2cover/blob/master/cover2cover.py
from lxml import etree as ET
import re
import os.path


def find_lines(j_package, filename):
    """Return all <line> elements for a given source file in a package."""
    lines = list()
    for sourcefile in j_package.iterfind("sourcefile"):
        if (
                sourcefile.attrib.get("name").split(".")[0]
                == os.path.basename(filename).split(".")[0]
        ):
            lines = lines + sourcefile.findall("line")
    return lines


def line_is_after(jm, start_line):
    return int(jm.attrib.get("line", 0)) > start_line


def method_lines(jmethod, jmethods, jlines):
    """Filter the lines from the given set of jlines that apply to the given jmethod."""
    start_line = int(jmethod.attrib.get("line", 0))
    larger = list(
        int(jm.attrib.get("line", 0))
        for jm in jmethods
        if line_is_after(jm, start_line)
    )
    end_line = min(larger) if len(larger) else 99999999

    for jline in jlines:
        if start_line <= int(jline.attrib["nr"]) < end_line:
            yield jline


def convert_lines(j_lines, into):
    """Convert the JaCoCo <line> elements into Cobertura <line> elements, add them under the given element."""
    c_lines = ET.SubElement(into, "lines")
    for jline in j_lines:
        mb = int(jline.attrib["mb"])
        cb = int(jline.attrib["cb"])
        ci = int(jline.attrib["ci"])

        cline = ET.SubElement(c_lines, "line")
        cline.set("number", jline.attrib["nr"])
        cline.set(
            "hits", "1" if ci > 0 else "0"
        )  # Probably not true but no way to know from JaCoCo XML file

        if mb + cb > 0:
            percentage = str(int(100 * (float(cb) / (float(cb) + float(mb))))) + "%"
            cline.set("branch", "true")
            cline.set(
                "condition-coverage",
                percentage + " (" + str(cb) + "/" + str(cb + mb) + ")",
            )

            cond = ET.SubElement(ET.SubElement(cline, "conditions"), "condition")
            cond.set("number", "0")
            cond.set("type", "jump")
            cond.set("coverage", percentage)
        else:
            cline.set("branch", "false")


def guess_filename(path_to_class, src_file_name):
    if src_file_name.endswith(".kt"):
        suffix = ".kt"
    else:
        suffix = ".java"
    m = re.match("([^$]*)", path_to_class)
    return (m.group(1) if m else path_to_class) + suffix


def add_counters(source, target):
    target.set("line-rate", counter(source, "LINE"))
    target.set("branch-rate", counter(source, "BRANCH"))
    target.set("complexity", counter(source, "COMPLEXITY", sum))


def fraction(covered, missed):
    return covered / (covered + missed)


def sum(covered, missed):
    return covered + missed


def counter(source, type, operation=fraction):
    cs = source.iterfind("counter")
    c = next((ct for ct in cs if ct.attrib.get("type") == type), None)

    if c is not None:
        covered = float(c.attrib["covered"])
        missed = float(c.attrib["missed"])

        return str(operation(covered, missed))
    else:
        return "0.0"


def convert_method(j_method, j_lines):
    c_method = ET.Element("method")
    c_method.set("name", j_method.attrib["name"])
    c_method.set("signature", j_method.attrib["desc"])

    add_counters(j_method, c_method)
    convert_lines(j_lines, c_method)

    return c_method


def convert_class(j_class, j_package):
    c_class = ET.Element("class")
    c_class.set("name", j_class.attrib["name"].replace("/", "."))

    # source file name can be None
    try:
        source_file_name = j_class.attrib["sourcefilename"]
    except KeyError:
        source_file_name = ""

    c_class.set(
        "filename",
        guess_filename(j_class.attrib["name"], source_file_name),
    )

    all_j_lines = list(find_lines(j_package, c_class.attrib["filename"]))

    # more than 8000 may causes mem issues
    if len(all_j_lines) > 8000:
        return c_class

    c_methods = ET.SubElement(c_class, "methods")
    all_j_methods = list(j_class.iterfind("method"))
    str_list = []
    for j_method in all_j_methods:
        j_method_lines = method_lines(j_method, all_j_methods, all_j_lines)
        each_node = convert_method(j_method, j_method_lines)
        str_list.append(ET.tostring(each_node, encoding="unicode"))

    for each in str_list:
        c_methods.append(ET.fromstring(each))

    add_counters(j_class, c_class)
    convert_lines(all_j_lines, c_class)
    return c_class


def convert_package(j_package):
    c_package = ET.Element("package")
    c_package.attrib["name"] = j_package.attrib["name"].replace("/", ".")

    c_classes = ET.SubElement(c_package, "classes")
    str_list = []
    for j_class in j_package.iterfind("class"):
        each_node = convert_class(j_class, j_package)
        str_list.append(ET.tostring(each_node, encoding="unicode"))

    for each in str_list:
        c_classes.append(ET.fromstring(each))

    add_counters(j_package, c_package)

    return c_package


def convert_root(source, target):
    try:
        ts = int(source.find("sessioninfo").attrib["start"]) / 1000
    except AttributeError:
        ts = -1
    target.set("timestamp", str(ts))

    packages = ET.SubElement(target, "packages")
    str_list = []
    for package in source.iterfind("package"):
        each_node = convert_package(package)
        str_list.append(ET.tostring(each_node, encoding="unicode"))

    for each in str_list:
        packages.append(ET.fromstring(each))

    add_counters(source, target)


def jacoco2cobertura(jacoco_string) -> str:
    root = ET.parse(jacoco_string).getroot()
    into = ET.Element("coverage")
    convert_root(root, into)
    output = f'<?xml version="1.0" ?>{ET.tostring(into, encoding="unicode")}'
    return output


# mem leak in lxml
# https://stackoverflow.com/a/49139904/10641498
# https://www.reddit.com/r/Python/comments/j0gl8t/psa_pythonlxml_memory_leaks_and_a_solution/
def destroy_tree(tree):
    root = tree
    node_tracker = {root: [0, None]}

    for node in root.iterdescendants():
        parent = node.getparent()
        node_tracker[node] = [node_tracker[parent][0] + 1, parent]

    node_tracker = sorted(
        [(depth, parent, child) for child, (depth, parent) in node_tracker.items()],
        key=lambda x: x[0],
        reverse=True,
    )

    for _, parent, child in node_tracker:
        if parent is None:
            break
        parent.remove(child)

    del tree



import fire
import json

from cobertura_parser.ext.lcov import lcov2cobertura
from cobertura_parser.loader import CoberturaLoader
from cobertura_parser.models.builtin import CoberturaStructure
from cobertura_parser.processor import CoberturaProcessor
from cobertura_parser.utils import time_measure
from cobertura_parser.ext.jacoco import jacoco2cobertura


class TerminalCli(object):
    def _cov(self, structure: CoberturaStructure) -> str:
        result = CoberturaProcessor.get_coverage(structure)
        result.lazy_calc()
        return result.json(
            exclude={
                "sources": ...,
                "line_rate": ...,
                "branch_rate": ...,
                "line_covered": ...,
                "line_valid": ...,
                "branches_covered": ...,
                "branches_valid": ...,
                "complexity": ...,
                "version": ...,
                "packages": {
                    "__all__": {
                        "classes": {
                            "__all__": {
                                "methods": {
                                    "__all__": {
                                        "branch_rate",
                                        "line_rate",
                                        "complexity",
                                    }
                                },
                                "branch_rate": ...,
                                "line_rate": ...,
                                "complexity": ...,
                                "lines": ...,
                            }
                        },
                        "branch_rate": ...,
                        "line_rate": ...,
                        "complexity": ...,
                    }
                },
            }
        )

    def _snapshot(self, structure: CoberturaStructure) -> str:
        result = CoberturaProcessor.get_code_snapshot_fat(structure)
        return result.json(
            exclude={
                "data": {
                    "sources": ...,
                    "line_rate": ...,
                    "branch_rate": ...,
                    "line_covered": ...,
                    "line_valid": ...,
                    "branches_covered": ...,
                    "branches_valid": ...,
                    "complexity": ...,
                    "version": ...,
                    "packages": {
                        "__all__": {
                            "classes": {
                                "__all__": {
                                    "methods": {
                                        "__all__": {
                                            "branch_rate",
                                            "line_rate",
                                            "complexity",
                                        }
                                    },
                                    "branch_rate": ...,
                                    "line_rate": ...,
                                    "complexity": ...,
                                }
                            },
                            "branch_rate": ...,
                            "line_rate": ...,
                            "complexity": ...,
                        }
                    },
                }
            }
        )

    def cov(self, from_file: str, to_file: str, dev: bool = None):
        with time_measure("cov", dev):
            structure: CoberturaStructure = CoberturaLoader.from_file(from_file)
            with open(to_file, "w") as f:
                f.write(self._cov(structure))

    def cov_from_jacoco(self, from_file: str, to_file: str, dev: bool = None):
        with time_measure("cov_from_jacoco", dev):
            structure: CoberturaStructure = CoberturaLoader.from_jacoco_file(from_file)
            with open(to_file, "w") as f:
                f.write(self._cov(structure))

    def snapshot(self, from_file: str, to_file: str, dev: bool = None):
        with time_measure("snapshot", dev):
            structure: CoberturaStructure = CoberturaLoader.from_file(from_file)
            with open(to_file, "w") as f:
                f.write(self._snapshot(structure))

    def snapshot_from_jacoco(self, from_file: str, to_file: str, dev: bool = None):
        with time_measure("snapshot_from_jacoco", dev):
            structure: CoberturaStructure = CoberturaLoader.from_jacoco_file(from_file)
            with open(to_file, "w") as f:
                f.write(self._snapshot(structure))

    def xml_from_jacoco(self, from_file: str, to_file: str, dev: bool = None):
        with time_measure("xml_from_jacoco", dev):
            with open(to_file, "w") as f:
                f.write(jacoco2cobertura(from_file))

    def xml_from_jacoco_to_json(self, from_file: str, to_file: str, dev: bool = None):
        with time_measure("xml_from_jacoco_to_json", dev):
            json_content: dict = CoberturaLoader.from_str(
                jacoco2cobertura(from_file), to_dict=True
            )
            with open(to_file, "w") as f:
                f.write(json.dumps(json_content))

    def data_from_lcov_to_json(self, from_file: str, to_file: str, dev: bool = None):
        with time_measure("data_from_lcov_to_json", dev):
            with open(from_file) as f:
                json_content: dict = CoberturaLoader.from_str(
                    lcov2cobertura(f.read()), to_dict=True
                )
            with open(to_file, "w") as f:
                f.write(json.dumps(json_content))

    def data_from_lcov_to_cov(self, from_file: str, to_file: str, dev: bool = None):
        with time_measure("data_from_lcov_to_cov", dev):
            with open(from_file) as f:
                structure: CoberturaStructure = CoberturaLoader.from_str(
                    lcov2cobertura(f.read())
                )
            with open(to_file, "w") as f:
                f.write(self._cov(structure))


def main():
    fire.Fire(TerminalCli)


if __name__ == "__main__":
    main()



import typing
from cobertura_parser.models.builtin import (
    CoberturaStructure,
    CoberturaStructureSlim,
)
from cobertura_parser.models.snapshot import CodeSnapshot, CodeSnapshotFat
from cobertura_parser.models.coverage import Coverage


class CoberturaProcessor(object):
    """
    processing some loaded data
    static class too, and these API should not change origin data
    """

    @classmethod
    def get_code_snapshot(
        cls, data: typing.Union[CoberturaStructure, CoberturaStructureSlim]
    ) -> CodeSnapshot:
        if isinstance(data, CoberturaStructure):
            return CodeSnapshot.from_normal(data)
        return CodeSnapshot.from_slim(data)

    @classmethod
    def get_code_snapshot_fat(
        cls, data: typing.Union[CoberturaStructure, CoberturaStructureSlim]
    ) -> CodeSnapshotFat:
        if isinstance(data, CoberturaStructure):
            return CodeSnapshotFat.from_normal(data)
        return CodeSnapshotFat.from_slim(data)

    @classmethod
    def get_coverage(
        cls, data: typing.Union[CoberturaStructure, CoberturaStructureSlim]
    ) -> Coverage:
        if isinstance(data, CoberturaStructure):
            return Coverage.from_normal(data)
        return Coverage.from_slim(data)



"""
comes from:
https://github.com/eriwen/lcov-to-cobertura-xml/blob/master/lcov_cobertura/lcov_cobertura.py
"""
import re
import sys
import os
import time
import subprocess
from xml.dom import minidom
from optparse import OptionParser

from distutils.spawn import find_executable

CPPFILT = "c++filt"
HAVE_CPPFILT = False

if find_executable(CPPFILT) is not None:
    HAVE_CPPFILT = True

VERSION = "1.6"
__all__ = ["LcovCobertura"]


class Demangler(object):
    def __init__(self):
        self.pipe = subprocess.Popen(
            CPPFILT, stdin=subprocess.PIPE, stdout=subprocess.PIPE
        )

    def demangle(self, name):
        newname = name + "\n"
        self.pipe.stdin.write(newname.encode("utf-8"))
        res = self.pipe.stdout.readline()
        return str(res.rstrip())


class LcovCobertura(object):
    """
    Converts code coverage report files in lcov format to Cobertura's XML
    report format so that CI servers like Jenkins can aggregate results and
    determine build stability etc.

    >>> from lcov_cobertura import LcovCobertura
    >>> LCOV_INPUT = 'your lcov input'
    >>> converter = LcovCobertura(LCOV_INPUT)
    >>> cobertura_xml = converter.convert()
    >>> print(cobertura_xml)
    """

    def __init__(self, lcov_data, base_dir=".", excludes=None, demangle=False):
        """
        Create a new :class:`LcovCobertura` object using the given `lcov_data`
        and `options`.

        :param lcov_data: Path to LCOV data file
        :type lcov_data: string
        :param base_dir: Path upon which to base all sources
        :type base_dir: string
        :param excludes: list of regexes to packages as excluded
        :type excludes: [string]
        :param demangle: whether to demangle function names using c++filt
        :type demangle: bool
        """

        if not excludes:
            excludes = []
        self.lcov_data = lcov_data
        self.base_dir = base_dir
        self.excludes = excludes
        if demangle:
            demangler = Demangler()
            self.format = demangler.demangle
        else:
            self.format = lambda x: x

    def convert(self):
        """
        Convert lcov file to cobertura XML using options from this instance.
        """
        coverage_data = self.parse()
        return self.generate_cobertura_xml(coverage_data)

    def parse(self):
        """
        Generate a data structure representing it that can be serialized in any
        logical format.
        """

        coverage_data = {
            "packages": {},
            "summary": {
                "lines-total": 0,
                "lines-covered": 0,
                "branches-total": 0,
                "branches-covered": 0,
            },
            "timestamp": str(int(time.time())),
        }
        package = None
        current_file = None
        file_lines_total = 0
        file_lines_covered = 0
        file_lines = {}
        file_methods = {}
        file_branches_total = 0
        file_branches_covered = 0

        for line in self.lcov_data.split("\n"):
            if line.strip() == "end_of_record":
                if current_file is not None:
                    package_dict = coverage_data["packages"][package]
                    package_dict["lines-total"] += file_lines_total
                    package_dict["lines-covered"] += file_lines_covered
                    package_dict["branches-total"] += file_branches_total
                    package_dict["branches-covered"] += file_branches_covered
                    file_dict = package_dict["classes"][current_file]
                    file_dict["lines-total"] = file_lines_total
                    file_dict["lines-covered"] = file_lines_covered
                    file_dict["lines"] = dict(file_lines)
                    file_dict["methods"] = dict(file_methods)
                    file_dict["branches-total"] = file_branches_total
                    file_dict["branches-covered"] = file_branches_covered
                    coverage_data["summary"]["lines-total"] += file_lines_total
                    coverage_data["summary"]["lines-covered"] += file_lines_covered
                    coverage_data["summary"]["branches-total"] += file_branches_total
                    coverage_data["summary"][
                        "branches-covered"
                    ] += file_branches_covered

            line_parts = line.split(":", 1)
            input_type = line_parts[0]

            if input_type == "SF":
                # Get file name
                file_name = line_parts[-1].strip()
                relative_file_name = os.path.relpath(file_name, self.base_dir)
                package = ".".join(relative_file_name.split(os.path.sep)[0:-1])
                class_name = ".".join(relative_file_name.split(os.path.sep))
                if package not in coverage_data["packages"]:
                    coverage_data["packages"][package] = {
                        "classes": {},
                        "lines-total": 0,
                        "lines-covered": 0,
                        "branches-total": 0,
                        "branches-covered": 0,
                    }
                coverage_data["packages"][package]["classes"][relative_file_name] = {
                    "name": class_name,
                    "lines": {},
                    "lines-total": 0,
                    "lines-covered": 0,
                    "branches-total": 0,
                    "branches-covered": 0,
                }
                package = package
                current_file = relative_file_name
                file_lines_total = 0
                file_lines_covered = 0
                file_lines.clear()
                file_methods.clear()
                file_branches_total = 0
                file_branches_covered = 0
            elif input_type == "DA":
                # DA:2,0
                (line_number, line_hits) = line_parts[-1].strip().split(",")
                line_number = int(line_number)
                if line_number not in file_lines:
                    file_lines[line_number] = {
                        "branch": "false",
                        "branches-total": 0,
                        "branches-covered": 0,
                    }
                file_lines[line_number]["hits"] = line_hits
                # Increment lines total/covered for class and package
                try:
                    if int(line_hits) > 0:
                        file_lines_covered += 1
                except:
                    pass
                file_lines_total += 1
            elif input_type == "BRDA":
                # BRDA:1,1,2,0
                (line_number, block_number, branch_number, branch_hits) = (
                    line_parts[-1].strip().split(",")
                )
                line_number = int(line_number)
                if line_number not in file_lines:
                    file_lines[line_number] = {
                        "branch": "true",
                        "branches-total": 0,
                        "branches-covered": 0,
                        "hits": 0,
                    }
                file_lines[line_number]["branch"] = "true"
                file_lines[line_number]["branches-total"] += 1
                file_branches_total += 1
                if branch_hits != "-" and int(branch_hits) > 0:
                    file_lines[line_number]["branches-covered"] += 1
                    file_branches_covered += 1
            elif input_type == "BRF":
                file_branches_total = int(line_parts[1])
            elif input_type == "BRH":
                file_branches_covered = int(line_parts[1])
            elif input_type == "FN":
                # FN:5,(anonymous_1)
                function_line, function_name = line_parts[-1].strip().split(",")
                file_methods[function_name] = [function_line, "0"]
            elif input_type == "FNDA":
                # FNDA:0,(anonymous_1)
                (function_hits, function_name) = line_parts[-1].strip().split(",")
                if function_name not in file_methods:
                    file_methods[function_name] = ["0", "0"]
                file_methods[function_name][-1] = function_hits

        # Exclude packages
        excluded = [
            x
            for x in coverage_data["packages"]
            for e in self.excludes
            if re.match(e, x)
        ]
        for package in excluded:
            del coverage_data["packages"][package]

        # Compute line coverage rates
        for package_data in list(coverage_data["packages"].values()):
            package_data["line-rate"] = self._percent(
                package_data["lines-total"], package_data["lines-covered"]
            )
            package_data["branch-rate"] = self._percent(
                package_data["branches-total"], package_data["branches-covered"]
            )

        return coverage_data

    def generate_cobertura_xml(self, coverage_data):
        """
        Given parsed coverage data, return a String cobertura XML representation.

        :param coverage_data: Nested dict representing coverage information.
        :type coverage_data: dict
        """

        dom_impl = minidom.getDOMImplementation()
        doctype = dom_impl.createDocumentType(
            "coverage", None, "http://cobertura.sourceforge.net/xml/coverage-04.dtd"
        )
        document = dom_impl.createDocument(None, "coverage", doctype)
        root = document.documentElement
        summary = coverage_data["summary"]
        self._attrs(
            root,
            {
                "branch-rate": self._percent(
                    summary["branches-total"], summary["branches-covered"]
                ),
                "branches-covered": str(summary["branches-covered"]),
                "branches-valid": str(summary["branches-total"]),
                "complexity": "0",
                "line-rate": self._percent(
                    summary["lines-total"], summary["lines-covered"]
                ),
                "lines-covered": str(summary["lines-covered"]),
                "lines-valid": str(summary["lines-total"]),
                "timestamp": coverage_data["timestamp"],
                "version": "1.0",
            },
        )

        sources = self._el(document, "sources", {})
        source = self._el(document, "source", {})
        source.appendChild(document.createTextNode(self.base_dir))
        sources.appendChild(source)

        root.appendChild(sources)

        packages_el = self._el(document, "packages", {})

        packages = coverage_data["packages"]
        for package_name, package_data in list(packages.items()):
            package_el = self._el(
                document,
                "package",
                {
                    "line-rate": package_data["line-rate"],
                    "branch-rate": package_data["branch-rate"],
                    "name": package_name,
                    "complexity": "0",
                },
            )
            classes_el = self._el(document, "classes", {})
            for class_name, class_data in list(package_data["classes"].items()):
                class_el = self._el(
                    document,
                    "class",
                    {
                        "branch-rate": self._percent(
                            class_data["branches-total"], class_data["branches-covered"]
                        ),
                        "complexity": "0",
                        "filename": class_name,
                        "line-rate": self._percent(
                            class_data["lines-total"], class_data["lines-covered"]
                        ),
                        "name": class_data["name"],
                    },
                )

                # Process methods
                methods_el = self._el(document, "methods", {})
                for method_name, (line, hits) in list(class_data["methods"].items()):
                    method_el = self._el(
                        document,
                        "method",
                        {
                            "name": self.format(method_name),
                            "signature": "",
                            "line-rate": "1.0" if int(hits) > 0 else "0.0",
                            "branch-rate": "1.0" if int(hits) > 0 else "0.0",
                        },
                    )
                    method_lines_el = self._el(document, "lines", {})
                    method_line_el = self._el(
                        document,
                        "line",
                        {
                            "hits": hits,
                            "number": line,
                            "branch": "false",
                        },
                    )
                    method_lines_el.appendChild(method_line_el)
                    method_el.appendChild(method_lines_el)
                    methods_el.appendChild(method_el)

                # Process lines
                lines_el = self._el(document, "lines", {})
                lines = list(class_data["lines"].keys())
                lines.sort()
                for line_number in lines:
                    line_el = self._el(
                        document,
                        "line",
                        {
                            "branch": class_data["lines"][line_number]["branch"],
                            "hits": str(class_data["lines"][line_number]["hits"]),
                            "number": str(line_number),
                        },
                    )
                    if class_data["lines"][line_number]["branch"] == "true":
                        total = int(class_data["lines"][line_number]["branches-total"])
                        covered = int(
                            class_data["lines"][line_number]["branches-covered"]
                        )
                        percentage = int((covered * 100.0) / total)
                        line_el.setAttribute(
                            "condition-coverage",
                            "{0}% ({1}/{2})".format(percentage, covered, total),
                        )
                    lines_el.appendChild(line_el)

                class_el.appendChild(methods_el)
                class_el.appendChild(lines_el)
                classes_el.appendChild(class_el)
            package_el.appendChild(classes_el)
            packages_el.appendChild(package_el)
        root.appendChild(packages_el)

        return document.toprettyxml()

    def _el(self, document, name, attrs):
        """
        Create an element within document with given name and attributes.

        :param document: Document element
        :type document: Document
        :param name: Element name
        :type name: string
        :param attrs: Attributes for element
        :type attrs: dict
        """
        return self._attrs(document.createElement(name), attrs)

    def _attrs(self, element, attrs):
        """
        Set attributes on given element.

        :param element: DOM Element
        :type element: Element
        :param attrs: Attributes for element
        :type attrs: dict
        """
        for attr, val in list(attrs.items()):
            element.setAttribute(attr, val)
        return element

    def _percent(self, lines_total, lines_covered):
        """
        Get the percentage of lines covered in the total, with formatting.

        :param lines_total: Total number of lines in given module
        :type lines_total: number
        :param lines_covered: Number of lines covered by tests in module
        :type lines_covered: number
        """

        if lines_total == 0:
            return "0.0"
        return str(float(float(lines_covered) / float(lines_total)))


def lcov2cobertura(data: str) -> str:
    converter = LcovCobertura(data)
    cobertura_xml = converter.convert()
    return cobertura_xml



from setuptools import setup, find_packages

__PROJECT_NAME__ = r"cobertura_parser"
__AUTHOR__ = r"williamfzc"
__AUTHOR_EMAIL__ = r"fengzc@vip.qq.com"
__VERSION__ = r"0.3.1"

setup(
    name=__PROJECT_NAME__,
    version=__VERSION__,
    author=__AUTHOR__,
    author_email=__AUTHOR_EMAIL__,
    packages=find_packages(),
    include_package_data=True,
    classifiers=[
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
    ],
    python_requires=">=3.6",
    install_requires=[
        "xmltodict",
        "planter",
        "pydantic",
        "fire",
        "lxml",
    ],
    entry_points={"console_scripts": ["cobertura_parser = cobertura_parser.cli:main"]},
)



"""
these models have same structure as origin cobertura dtd
see http://cobertura.sourceforge.net/xml/coverage-04.dtd
"""
from pydantic import BaseModel
import typing
from cobertura_parser.utils import unused_dict_to_list

TYPE_ORIGIN_CONDITIONS = typing.Dict[
    str, typing.Union["CoberturaCondition", typing.List["CoberturaCondition"]]
]
TYPE_ORIGIN_LINES = typing.Optional[
    typing.Dict[str, typing.Union["CoberturaLine", typing.List["CoberturaLine"]]]
]
TYPE_ORIGIN_METHODS = typing.Optional[
    typing.Dict[str, typing.Union["CoberturaMethod", typing.List["CoberturaMethod"]]]
]
TYPE_ORIGIN_KLASSES = typing.Dict[
    str, typing.Union["CoberturaKlass", typing.List["CoberturaKlass"]]
]
TYPE_ORIGIN_PACKAGES = typing.Dict[
    str, typing.Union["CoberturaPackage", typing.List["CoberturaPackage"]]
]


class CoberturaCondition(BaseModel):
    number: int
    type: str
    coverage: str

    class Config:
        allow_population_by_field_name = True
        fields = {
            "number": {"alias": "@number"},
            "type": {"alias": "@type"},
            "coverage": {"alias": "@coverage"},
        }


class CoberturaLine(BaseModel):
    number: int
    hits: int
    # can not be `bool` because it can not be cast to xml format back
    branch: str
    condition_coverage: str = None

    conditions: TYPE_ORIGIN_CONDITIONS = None

    class Config:
        allow_population_by_field_name = True
        fields = {
            "number": {"alias": "@number"},
            "hits": {"alias": "@hits"},
            "branch": {"alias": "@branch"},
            "condition_coverage": {"alias": "@condition-coverage"},
        }

    def is_hit(self) -> bool:
        return bool(self.hits)

    def is_in_branch(self) -> bool:
        # todo: is it correct??
        return self.branch == "true"


class CoberturaMethod(BaseModel):
    name: str
    signature: str
    line_rate: float
    branch_rate: float
    complexity: float = None

    # 3 types
    # - None
    # - Dict (only one element
    # - List (more than one
    lines: TYPE_ORIGIN_LINES = dict()

    class Config:
        allow_population_by_field_name = True
        fields = {
            "name": {"alias": "@name"},
            "signature": {"alias": "@signature"},
            "line_rate": {"alias": "@line-rate"},
            "branch_rate": {"alias": "@branch-rate"},
            "complexity": {"alias": "@complexity"},
        }

    def get_line_list(self) -> typing.List[CoberturaLine]:
        return unused_dict_to_list(self.lines)


class CoberturaKlass(BaseModel):
    name: str
    filename: str
    line_rate: float
    branch_rate: float
    complexity: float = None

    methods: TYPE_ORIGIN_METHODS = dict()
    lines: TYPE_ORIGIN_LINES = dict()

    class Config:
        allow_population_by_field_name = True
        fields = {
            "name": {"alias": "@name"},
            "filename": {"alias": "@filename"},
            "line_rate": {"alias": "@line-rate"},
            "branch_rate": {"alias": "@branch-rate"},
            "complexity": {"alias": "@complexity"},
        }

    def get_method_list(self) -> typing.List[CoberturaMethod]:
        return unused_dict_to_list(self.methods)

    def get_line_list(self) -> typing.List[CoberturaLine]:
        return unused_dict_to_list(self.lines)


class CoberturaPackage(BaseModel):
    name: str
    line_rate: float
    branch_rate: float
    complexity: float = None
    classes: TYPE_ORIGIN_KLASSES = dict()

    class Config:
        allow_population_by_field_name = True
        fields = {
            "name": {"alias": "@name"},
            "line_rate": {"alias": "@line-rate"},
            "branch_rate": {"alias": "@branch-rate"},
            "complexity": {"alias": "@complexity"},
        }

    def get_class_list(self) -> typing.List[CoberturaKlass]:
        return unused_dict_to_list(self.classes)


class CoberturaCoverage(BaseModel):
    sources: typing.Dict[str, typing.Union[str, typing.List[str]]] = None
    packages: TYPE_ORIGIN_PACKAGES = None

    # attrs
    line_rate: float
    branch_rate: float
    line_covered: int = None
    line_valid: int = None
    branches_covered: int = None
    branches_valid: int = None
    complexity: float = None
    version: float = -1.0
    timestamp: float = -1.0

    class Config:
        allow_population_by_field_name = True
        fields = {
            "line_rate": {"alias": "@line-rate"},
            "branch_rate": {"alias": "@branch-rate"},
            "line_covered": {"alias": "@lines-covered"},
            "line_valid": {"alias": "@lines-valid"},
            "branches_covered": {"alias": "@branches-covered"},
            "branches_valid": {"alias": "@branches-valid"},
            "complexity": {"alias": "@complexity"},
            "version": {"alias": "@version"},
            "timestamp": {"alias": "@timestamp"},
        }

    def get_package_list(self) -> typing.List[CoberturaPackage]:
        return unused_dict_to_list(self.packages)

    def slim(self) -> "CoberturaStructureSlim":
        data = CoberturaCoverage(**self.dict())
        for each_pkg in data.get_package_list():
            for each_kls in each_pkg.get_class_list():
                for each_method in each_kls.get_method_list():
                    each_method.lines = each_method.get_line_list()
                each_kls.methods = each_kls.get_method_list()
                each_kls.lines = each_kls.get_line_list()
            each_pkg.classes = each_pkg.get_class_list()
        data.packages = data.get_package_list()
        return CoberturaStructureSlim(**data.dict())


class CoberturaLineSlim(CoberturaLine):
    pass


class CoberturaMethodSlim(CoberturaMethod):
    lines: typing.List[CoberturaLine] = None

    def get_line_list(self) -> typing.List[CoberturaLine]:
        raise NotImplementedError

    def is_hit(self) -> bool:
        return any((each.is_hit() for each in self.lines))


class CoberturaKlassSlim(CoberturaKlass):
    methods: typing.List[CoberturaMethodSlim] = None
    lines: typing.List[CoberturaLineSlim] = None

    def get_method_list(self) -> typing.List[CoberturaMethod]:
        raise NotImplementedError

    def get_line_list(self) -> typing.List[CoberturaLine]:
        raise NotImplementedError

    def is_hit(self) -> bool:
        return any((each.is_hit() for each in self.methods))


class CoberturaPackageSlim(CoberturaPackage):
    classes: typing.List[CoberturaKlassSlim] = None

    def get_class_list(self) -> typing.List[CoberturaKlass]:
        raise NotImplementedError

    def is_hit(self) -> bool:
        return any((each.is_hit() for each in self.classes))


class CoberturaStructureSlim(CoberturaCoverage):
    packages: typing.List[CoberturaPackageSlim] = None

    def get_package_list(self) -> typing.List[CoberturaPackage]:
        raise NotImplementedError


class CoberturaStructure(BaseModel):
    coverage: CoberturaCoverage

    def to_origin(self) -> dict:
        return self.dict(by_alias=True, exclude_defaults=True)

    def slim(self) -> CoberturaStructureSlim:
        return self.coverage.slim()



import typing
from pydantic import validator
from cobertura_parser.models.builtin import (
    CoberturaMethodSlim,
    CoberturaKlassSlim,
    CoberturaPackageSlim,
    CoberturaStructureSlim,
    CoberturaStructure,
    CoberturaLineSlim,
)


class CoverageMethod(CoberturaMethodSlim):
    valid_lines: int = -1
    valid_branches: int = -1

    def lazy_calc(self):
        self.valid_lines = len(self.lines)
        self.valid_branches = len([each for each in self.lines if each.is_in_branch()])
        self.lines = [each for each in self.lines if each.is_hit()]


class CoverageKlass(CoberturaKlassSlim):
    methods: typing.List[CoverageMethod]

    @validator("methods")
    def remove_no_hit_methods(cls, methods: typing.List[CoverageMethod]):
        return [each for each in methods if each.is_hit()]

    def lazy_calc(self):
        for each in self.methods:
            each.lazy_calc()


class CoveragePackage(CoberturaPackageSlim):
    classes: typing.List[CoverageKlass]

    @validator("classes")
    def remove_no_hit_classes(cls, classes: typing.List[CoverageKlass]):
        return [each for each in classes if each.is_hit()]

    def lazy_calc(self):
        for each in self.classes:
            each.lazy_calc()


class Coverage(CoberturaStructureSlim):
    packages: typing.List[CoveragePackage] = None

    @validator("packages")
    def remove_no_hit_package(cls, packages: typing.List[CoveragePackage]):
        return [each for each in packages if each.is_hit()]

    @classmethod
    def from_slim(cls, slim_data: CoberturaStructureSlim) -> "Coverage":
        return cls(**slim_data.dict())

    @classmethod
    def from_normal(cls, data: CoberturaStructure) -> "Coverage":
        return cls.from_slim(data.slim())

    def lazy_calc(self):
        for each in self.packages:
            each.lazy_calc()
